@@ -4,7 +4,6 @@
 
 package Net::LDAP::Filter;
 
-use Net::LDAP::BER ();
 use strict;
 use vars qw($VERSION);
 
@@ -48,7 +47,7 @@ sub new {
   my $self = shift;
   my $class = ref($self) || $self;
   
-  my $me = bless [], $class;
+  my $me = bless {}, $class;
 
   if (@_) {
     $me->parse(shift) or
@@ -57,19 +56,17 @@ sub new {
   $me;
 }
 
-my($Attr)  = qw{  [-;.:\d\w]*[-;\d\w] };
-my($Op)    = qw{  [:~<>]?=            };
-my($Value) = qw{  (?:\\.|[^\\()]+)*   };
+my $Attr  = '[-;.:\d\w]*[-;\d\w]';
 
 my %Op = qw(
-  &   FILTER_AND
-  |   FILTER_OR
-  !   FILTER_NOT
-  =   FILTER_EQ
-  ~=  FILTER_APPROX
-  >=  FILTER_GE
-  <=  FILTER_LE
-  :=  FILTER_EXTENSIBLE
+  &   and
+  |   or
+  !   not
+  =   equalityMatch
+  ~=  approxMatch
+  >=  greaterOrEqual
+  <=  lessOrEqual
+  :=  extensibleMatch
 );
 
 my %Rop = reverse %Op;
@@ -91,7 +88,9 @@ sub _unescape {
   $_[0];
 }
 
-sub _escape { $_[0] =~ s/([\\\(\)\*\0])/sprintf("\\%02x",ord($1))/sge; }
+my %ch = split(/\s+/, '( \\( ) \\) \\ \\\\ * \\*');
+
+sub _escape { (my $t = $_[0]) =~ s/([\\\(\)\*\0-\37])/$ch{$1}||sprintf("\\%02x",ord($1))/sge; $t }
 
 sub _encode {
   my($attr,$op,$val) = @_;
@@ -107,14 +106,14 @@ sub _encode {
     }
     my($type,$dn,$rule) = ($1,$2,$4);
 
-    return (
-      FILTER_EXTENSIBLE => [
-	OPTIONAL => [ EXTENSIBLE_RULE => $rule ],
-	OPTIONAL => [ EXTENSIBLE_TYPE => $type ],
-	EXTENSIBLE_VALUE => _unescape($val),
-	EXTENSIBLE_DN    => $dn
-      ]
-    );
+    return ( {
+      extensibleMatch => {
+	matchingRule => $rule,
+	type         => $type,
+	matchValue   => _unescape($val), 
+	dnAttributes => $dn ? 1 : 0
+      }
+    });
   }
 
   # If the op is = and contains one or more * not
@@ -122,34 +121,33 @@ sub _encode {
 
   if ($op eq '=' && $val =~ /^(?:(?:\\\*|[^*])*)\*/o) {
     my $n = [];
-    my $type = 'SUBSTR_INITIAL';
+    my $type = 'initial';
 
     while ($val =~ s/^((?:\\\*|[^*])*)\*+//) {
-      push(@$n, $type, _unescape("$1"))         # $1 is readonly, copy it
+      push(@$n, { $type, _unescape("$1") })         # $1 is readonly, copy it
 	if length $1;
 
-      $type = 'SUBSTR_ANY';
+      $type = 'any';
     }
 
-    push(@$n, 'SUBSTR_FINAL', _unescape($val))
+    push(@$n, { 'final', _unescape($val) })
       if length $val;
 
-    return (
-      FILTER_SUBSTRS => [
-	STRING   => $attr,
-	SEQUENCE => $n
-      ]
-    );
+    return ({
+      substrings => {
+	type       => $attr,
+	substrings => $n
+      }
+    });
   }
 
   # Well we must have an operator and no un-escaped *'s on the RHS
 
-  return (
-    $Op{$op} => [
-      STRING => $attr,
-      STRING => _unescape($val)
-    ]
-  );
+  return {
+    $Op{$op} => {
+      attributeDesc => $attr, assertionValue =>  _unescape($val)
+    }
+  };
 }
 
 sub parse {
@@ -157,10 +155,10 @@ sub parse {
   my $filter = shift;
 
   my @stack = ();   # stack
-  my $cur   = $self;
+  my $cur   = [];
+  my $op;
 
   undef $ErrStr;
-  @$cur = ();
 
   # Algorithm depends on /^\(/;
   $filter =~ s/^\s*//;
@@ -172,36 +170,40 @@ sub parse {
 
     # Process the start of  (& (...)(...))
 
-    if ($filter =~ s/^\(\s*([&|!])\s*//) {
-      my $n = [];                   # new list to hold filter elements
-
-      push(@$cur, $Op{$1}, $n);
-      push(@stack,$cur);            # push current list on the stack
-
-      $cur = $n;
+    if ($filter =~ s/^\(\s*([&!|])\s*//) {
+      push @stack, [$op,$cur];
+      $op = $1;
+      $cur = [];
       next;
     }
 
     # Process the end of  (& (...)(...))
 
-    if ($filter =~ s/^\)\s*//o) {
-      $cur = pop @stack;
-      last unless @stack;
-      next;
+    elsif ($filter =~ s/^\)\s*//o) {
+      my($myop,$mydata) = ($op,$cur);
+      ($op,$cur) = @{ pop @stack };
+	# Need to do more checking here
+      push @$cur, { $Op{$myop} => $myop eq '!' ? $mydata->[0] : $mydata };
+      next if @stack;
     }
     
     # present is a special case (attr=*)
 
-    if ($filter =~ s/^\(\s*($Attr)=\*\)\s*//o) {
-      push(@$cur, FILTER_PRESENT => $1);
-      next;
+    elsif ($filter =~ s/^\(\s*($Attr)=\*\)\s*//o) {
+      push(@$cur, { present => $1 } );
+      next if @stack;
     }
 
     # process (attr op string)
 
-    if ($filter =~ s/^\(\s*($Attr)\s*($Op)($Value)\)\s*//o) {
+    elsif ($filter =~ s/^\(\s*
+                        ($Attr)\s*
+                        ([:~<>]?=)
+                        ((?:\\.|[^\\()]+)*)
+                        \)\s*
+                       //xo) {
       push(@$cur, _encode($1,$2,$3));
-      next;
+      next if @stack;
     }
 
     # If we get here then there is an error in the filter string
@@ -215,19 +217,11 @@ sub parse {
     return undef;
   }
 
+  %$self = %{$cur->[0]};
+
   $self;
 }
 
-sub ber {
-  my $self = shift;
-  my $ber = new Net::LDAP::BER();
-
-  return undef
-    unless $ber->encode( @$self );
-
-  $ber;
-}
-
 sub print {
   my $self = shift;
   no strict 'refs'; # select may return a GLOB name
@@ -236,84 +230,35 @@ sub print {
   print $fh $self->as_string,"\n";
 }
 
-sub as_string { _string(@{$_[0]}) }
+sub as_string { _string(%{$_[0]}) }
 
 sub _string {    # prints things of the form (<op> (<list>) ... )
-  my @self = @_;
   my $i;
   my $str = "";
 
-  for ($i=0; $i <= $#self; $i+=2) {  # List of ( operator, list ... )
-    my $op = $Rop{$self[$i]} || '';
-    if ($op =~ /^[&!|]$/) {  
-      $str .= "($op" . _string(@{$self[$i+1]}) . ")";
-    } else {
-      $str .= _string_infix(@self[$i,$i+1],$op);
-    }
-  }
-  $str;
-}
-
-sub _string_infix {    #  prints infix items of the form ( <attrib> <op> <val> )
-  my($tag, $items, $op) = @_;
-
-  # An EXTENSIBLE match
-
-  if ($op eq ':=') {
-    my($rule,$type,$val,$dn) = ($items->[1][1],$items->[3][1],$items->[5],$items->[7]);
-
-    $val =~ s/([\\\(\)\*\0])/sprintf("\\%02x",ord($1))/sge;
-
-    return join("",
-         '(',
-	   ($type ? ($type)     : ()),
-	   ($dn   ? (':dn')     : ()),
-	   ($rule ? (':',$rule) : ()),
-	   ':=',
-	   $val,
-	 ')');
-
+  for ($_[0]) {
+    /^and/ and return "(&" . join("", map { _string(%$_) } @{$_[1]}) . ")";
+    /^or/  and return "(|" . join("", map { _string(%$_) } @{$_[1]}) . ")";
+    /^not/ and return "(!" . _string(%{$_[1]}) . ")";
+    /^present/ and return "($_[1]=*)";
+    /^(equalityMatch|greaterOrEqual|lessOrEqual|approxMatch)/
+      and return "(" . $_[1]->{attributeDesc} . $Rop{$1} . _escape($_[1]->{assertionValue})  .")";
+    /^substrings/ and do {
+      my $str = join("*", "",map { _escape($_) } map { values %$_ } @{$_[1]->{substrings}});
+      $str =~ s/^.// if exists $_[1]->{substrings}[0]{initial};
+      $str .= '*' unless exists $_[1]->{substrings}[-1]{final};
+      return "($_[1]->{type}=$str)";
+    };
+    /^extensibleMatch/ and do {
+      my $str = "($_[1]->{type}";
+      $str .= ":dn" if $_[1]->{dnAttributes};
+      $str .= ":$_[1]->{matchingRule}" if defined $_[1]->{matchingRule};
+      $str .= ":=" . _escape($_[1]->{matchValue}) . ")";
+      return $str;
+    };
   }
 
-  # ~= >= <= or simple =
-
-  if (length $op) {
-    my $val = $items->[3];
-
-    $val =~ s/([\\\(\)\*\0])/sprintf("\\%02x",ord($1))/sge;
-
-    return "($items->[1]$op$val)";
-  }
-
-  # PRESENT
-
-  if ($tag eq 'FILTER_PRESENT') {
-    return "($items=*)";
-  }
-
-  # SUBSTRS
-
-  if ($tag eq 'FILTER_SUBSTRS') {
-    my @bits = ( '(', $items->[1], '=');
-    my $substrs = $items->[3];
-    my $substr;
-
-    push(@bits, '*')
-        if $substrs->[0] ne 'SUBSTR_INITIAL';
-
-    for( $substr=0; $substr < $#{$substrs}; $substr += 2) {
-      my $val = $substrs->[$substr+1];
-
-      $val =~ s/([\\\(\)\*\0])/sprintf("\\%02x",ord($1))/sge;
-
-      push(@bits,
-           $val,
-	   $substrs->[$substr] ne 'SUBSTR_FINAL' ? '*' : ());
-    }
-    return join("",@bits,")");
-  }
-
-  return "";
+  die "Internal error $_[0]";
 }
 
 1;
